@@ -11,169 +11,96 @@
 #include "stat_cache.h"
 #include "chunk.h"
 #include "http_chunk.h"
-#include "http_date.h"
 #include "http_range.h"
 
 #include "plugin.h"
+#include "plugins.h"
 
 #include <sys/types.h>
-#include <sys/stat.h>
+#include "sys-stat.h"
 #include "sys-time.h"
 
-#include <limits.h>
 #include <errno.h>
-#include <stdlib.h>
 #include <string.h>
 
 
-int
-http_response_omit_header (request_st * const r, const data_string * const ds)
+static stat_cache_entry *
+http_response_physical_pathinfo (request_st * const r)
 {
-    const size_t klen = buffer_clen(&ds->key);
-    if (klen == sizeof("X-Sendfile")-1
-        && buffer_eq_icase_ssn(ds->key.ptr, CONST_STR_LEN("X-Sendfile")))
-        return 1;
-    if (klen >= sizeof("X-LIGHTTPD-")-1
-        && buffer_eq_icase_ssn(ds->key.ptr, CONST_STR_LEN("X-LIGHTTPD-"))) {
-        if (klen == sizeof("X-LIGHTTPD-KBytes-per-second")-1
-            && buffer_eq_icase_ssn(ds->key.ptr+sizeof("X-LIGHTTPD-")-1,
-                                   CONST_STR_LEN("KBytes-per-second"))) {
-            /* "X-LIGHTTPD-KBytes-per-second" */
-            off_t limit = strtol(ds->value.ptr, NULL, 10) << 10; /*(*=1024)*/
-            if (limit > 0
-                && (limit < r->conf.bytes_per_second
-                    || 0 == r->conf.bytes_per_second)) {
-                r->conf.bytes_per_second = limit;
-            }
+    /* Caller must already have checked full path does not exist in filesystem*/
+
+    char *pathinfo = r->physical.path.ptr
+                   + buffer_clen(&r->physical.basedir)
+                   - (buffer_has_pathsep_suffix(&r->physical.basedir));
+    if ('/' != *pathinfo)
+        pathinfo = NULL;
+    else if (pathinfo == r->physical.path.ptr) /*(basedir is "/")*/
+        pathinfo = strchr(pathinfo+1, '/');
+    /* Note: basedir might be "/" (containers) and basedir should not be empty.
+     * basedir in config is allowed to end with '/', especially for basedir "/".
+     * Current implmentation below requires that pathinfo follow a regular file
+     * (S_ISREG()) which is why if pathinfo matches beginning of path, pathinfo
+     * is stepped to next path component.
+     *   https://redmine.lighttpd.net/issues/2911
+     */
+
+    stat_cache_entry *sce = NULL;
+    const uint32_t pathused = r->physical.path.used;
+    for (char *pprev = pathinfo;
+         pathinfo;
+         pprev = pathinfo, pathinfo = strchr(pathinfo+1, '/')) {
+        /*(temporarily modify r->physical.path in-place)*/
+        r->physical.path.used = pathinfo - r->physical.path.ptr + 1;
+        *pathinfo = '\0';
+        stat_cache_entry * const nsce = stat_cache_get_entry(&r->physical.path);
+        *pathinfo = '/';
+        r->physical.path.used = pathused;
+        if (NULL == nsce) {
+            pathinfo = pathinfo != pprev ? pprev : NULL;
+            break;
         }
-        return 1;
+        sce = nsce;
+        if (!S_ISDIR(sce->st.st_mode)) break;
     }
-    return 0;
-}
 
+    /* Note: historical lighttpd behavior checks S_ISREG(), permitting
+     * pathinfo only on regular files, not dirs or special files.
+     *
+     * Were this code to be extended to permit pathinfo following a dir, the
+     * trailing slash indicating dir would have to be duplicated to start
+     * pathinfo and would need to be special-cased in the two calls to
+     * buffer_truncate() below.  Additionally, basedir "/" and entire rest of
+     * path as pathinfo would have to be special-cased here before returning,
+     * including sce = stat_cache_get_entry() "/".  However, supporting pathinfo
+     * on dirs -- at this point in lighttpd request processing -- would have the
+     * effect of every request for a non-existent file being pathinfo on a dir
+     * (assuming "/" exists), instead of returning the traditional 404 Not Found
+     * in such cases.  Fully virtual paths are handled elsewhere,
+     * e.g. with gw_backend "check-local" => "disable" in lighttpd.conf */
 
-__attribute_cold__
-static void
-http_response_write_header_partial_1xx (request_st * const r, buffer * const b)
-{
-    /* take data in con->write_queue and move into b
-     * (to be sent prior to final response headers in r->write_queue) */
-    connection * const con = r->con;
-    /*assert(&r->write_queue != con->write_queue);*/
-    chunkqueue * const cq = con->write_queue;
-    con->write_queue = &r->write_queue;
+    if (NULL == pathinfo || !S_ISREG(sce->st.st_mode))
+        return NULL;
 
-    /*assert(0 == buffer_clen(b));*//*expect empty buffer from caller*/
-    uint32_t len = (uint32_t)chunkqueue_length(cq);
-    /*(expecting MEM_CHUNK(s), so not expecting error reading files)*/
-    if (chunkqueue_read_data(cq, buffer_string_prepare_append(b, len),
-                             len, r->conf.errh) < 0)
-        len = 0;
-    buffer_truncate(b, len);/*expect initial empty buffer from caller*/
-    chunkqueue_free(cq);
-}
+    /* pathinfo */
+    size_t len = r->physical.path.ptr+pathused-1-pathinfo, reqlen;
+    const char * const ptr =
+       (r->conf.force_lowercase_filenames
+        && len <= (reqlen = buffer_clen(&r->target))
+        && buffer_eq_icase_ssn(r->target.ptr + reqlen - len, pathinfo, len))
+        /* attempt to preserve case-insensitive PATH_INFO
+         * (works in common case where mod_alias, mod_magnet, and other modules
+         *  have not modified the PATH_INFO portion of request URI, or did so
+         *  with exactly the PATH_INFO desired) */
+      ? r->target.ptr + reqlen - len
+      : pathinfo;
+    buffer_copy_string_len(&r->pathinfo, ptr, len);
 
+    /* remove pathinfo from paths */
+    buffer_truncate(&r->uri.path, buffer_clen(&r->uri.path) - len);
+    buffer_truncate(&r->physical.path,
+                    (uint32_t)(pathinfo - r->physical.path.ptr));
 
-void
-http_response_write_header (request_st * const r)
-{
-	/* disable keep-alive if requested */
-
-	r->con->keep_alive_idle = r->conf.max_keep_alive_idle;
-	if (__builtin_expect( (0 == r->conf.max_keep_alive_idle), 0)
-	    || r->con->request_count > r->conf.max_keep_alive_requests) {
-		r->keep_alive = 0;
-	} else if (0 != r->reqbody_length
-		   && r->reqbody_length != r->reqbody_queue.bytes_in
-		   && (NULL == r->handler_module
-		       || 0 == (r->conf.stream_request_body
-		                & (FDEVENT_STREAM_REQUEST
-		                   | FDEVENT_STREAM_REQUEST_BUFMIN)))) {
-		r->keep_alive = 0;
-	}
-
-	if (light_btst(r->resp_htags, HTTP_HEADER_UPGRADE)
-	    && r->http_version == HTTP_VERSION_1_1) {
-		http_header_response_set(r, HTTP_HEADER_CONNECTION, CONST_STR_LEN("Connection"), CONST_STR_LEN("upgrade"));
-	} else if (r->keep_alive <= 0) {
-		http_header_response_set(r, HTTP_HEADER_CONNECTION, CONST_STR_LEN("Connection"), CONST_STR_LEN("close"));
-	} else if (r->http_version == HTTP_VERSION_1_0) {/*(&& r->keep_alive > 0)*/
-		http_header_response_set(r, HTTP_HEADER_CONNECTION, CONST_STR_LEN("Connection"), CONST_STR_LEN("keep-alive"));
-	}
-
-	if (304 == r->http_status
-	    && light_btst(r->resp_htags, HTTP_HEADER_CONTENT_ENCODING)) {
-		http_header_response_unset(r, HTTP_HEADER_CONTENT_ENCODING, CONST_STR_LEN("Content-Encoding"));
-	}
-
-	chunkqueue * const cq = &r->write_queue;
-	buffer * const b = chunkqueue_prepend_buffer_open(cq);
-
-	if (cq != r->con->write_queue)
-		http_response_write_header_partial_1xx(r, b);
-
-	buffer_append_string_len(b,
-	                         (r->http_version == HTTP_VERSION_1_1)
-	                           ? "HTTP/1.1 "
-	                           : "HTTP/1.0 ",
-	                         sizeof("HTTP/1.1 ")-1);
-	http_status_append(b, r->http_status);
-
-	/* add all headers */
-	for (size_t i = 0, used = r->resp_headers.used; i < used; ++i) {
-		const data_string * const ds = (data_string *)r->resp_headers.data[i];
-		const uint32_t klen = buffer_clen(&ds->key);
-		const uint32_t vlen = buffer_clen(&ds->value);
-		if (__builtin_expect( (0 == klen), 0)) continue;
-		if (__builtin_expect( (0 == vlen), 0)) continue;
-		if ((ds->key.ptr[0] & 0xdf) == 'X' && http_response_omit_header(r, ds))
-			continue;
-		char * restrict s = buffer_extend(b, klen+vlen+4);
-		s[0] = '\r';
-		s[1] = '\n';
-		memcpy(s+2, ds->key.ptr, klen);
-		s += 2+klen;
-		s[0] = ':';
-		s[1] = ' ';
-		memcpy(s+2, ds->value.ptr, vlen);
-	}
-
-	if (!light_btst(r->resp_htags, HTTP_HEADER_DATE)) {
-		/* HTTP/1.1 and later requires a Date: header */
-		/* "\r\nDate: " 8-chars + 30-chars "%a, %d %b %Y %T GMT" + '\0' */
-		static unix_time64_t tlast = 0;
-		static char tstr[40] = "\r\nDate: ";
-
-		/* cache the generated timestamp */
-		const unix_time64_t cur_ts = log_epoch_secs;
-		if (__builtin_expect ( (tlast != cur_ts), 0))
-			http_date_time_to_str(tstr+8, sizeof(tstr)-8, (tlast = cur_ts));
-
-		buffer_append_string_len(b, tstr, 37);
-	}
-
-	if (!light_btst(r->resp_htags, HTTP_HEADER_SERVER) && r->conf.server_tag)
-		buffer_append_str2(b, CONST_STR_LEN("\r\nServer: "),
-		                      BUF_PTR_LEN(r->conf.server_tag));
-
-	buffer_append_string_len(b, CONST_STR_LEN("\r\n\r\n"));
-
-	r->resp_header_len = buffer_clen(b);
-
-	if (r->conf.log_response_header) {
-		log_error_multiline(r->conf.errh, __FILE__, __LINE__,
-		                    BUF_PTR_LEN(b), "fd:%d resp: ", r->con->fd);
-	}
-
-	chunkqueue_prepend_buffer_commit(cq);
-
-	/*(optimization to use fewer syscalls to send a small response)*/
-	off_t cqlen;
-	if (r->resp_body_finished
-	    && light_btst(r->resp_htags, HTTP_HEADER_CONTENT_LENGTH)
-	    && (cqlen = chunkqueue_length(cq) - r->resp_header_len) > 0
-	    && cqlen < 16384)
-		chunkqueue_small_resp_optim(cq);
+    return sce;
 }
 
 
@@ -209,14 +136,26 @@ static handler_t http_response_physical_path_check(request_st * const r) {
 			break;
 		case EACCES:
 			return http_response_physical_path_error(r, 403, NULL);
-		case ENAMETOOLONG:
-			/* file name to be read was too long. return 404 */
 		case ENOENT:
 			if (r->http_method == HTTP_METHOD_OPTIONS
 			    && light_btst(r->resp_htags, HTTP_HEADER_ALLOW)) {
 				r->http_status = 200;
 				return HANDLER_FINISHED;
 			}
+		  #ifdef _WIN32
+			/* _WIN32 returns ENOENT instead of ENOTDIR for PATH_INFO */
+			break;
+		  #else
+			__attribute_fallthrough__
+		  #endif
+		case ENAMETOOLONG:
+			/* file name to be read was too long. return 404 */
+			/* Note: URIs can be very long and this initial check on the
+			 * entire path imposes limits use of pathinfo in URIs to
+			 * (typically) 255 byte path segments and (typically) 4k total len,
+			 * though these limits can be avoided by configuring some modules
+			 * to use virtual paths and to skip the filesystem check,
+			 * e.g. w/ gw_backend "check-local" => "disable" in lighttpd.conf */
 			return http_response_physical_path_error(r, 404, NULL);
 		default:
 			/* we have no idea what happened. let's tell the user so. */
@@ -225,65 +164,9 @@ static handler_t http_response_physical_path_check(request_st * const r) {
 
 		/* not found, perhaps PATHINFO */
 
-		char *pathinfo;
-		{
-			/*(might check at startup that s->document_root does not end in '/')*/
-			size_t len = buffer_clen(&r->physical.basedir)
-			           - (buffer_has_pathsep_suffix(&r->physical.basedir));
-			pathinfo = r->physical.path.ptr + len;
-			if ('/' != *pathinfo) {
-				pathinfo = NULL;
-			}
-			else if (pathinfo == r->physical.path.ptr) { /*(basedir is "/")*/
-				pathinfo = strchr(pathinfo+1, '/');
-			}
-		}
-
-		const uint32_t pathused = r->physical.path.used;
-		for (char *pprev = pathinfo; pathinfo; pprev = pathinfo, pathinfo = strchr(pathinfo+1, '/')) {
-			/*(temporarily modify r->physical.path in-place)*/
-			r->physical.path.used = pathinfo - r->physical.path.ptr + 1;
-			*pathinfo = '\0';
-			stat_cache_entry * const nsce = stat_cache_get_entry(&r->physical.path);
-			*pathinfo = '/';
-			r->physical.path.used = pathused;
-			if (NULL == nsce) {
-				pathinfo = pathinfo != pprev ? pprev : NULL;
-				break;
-			}
-			sce = nsce;
-			if (!S_ISDIR(sce->st.st_mode)) break;
-		}
-
-		if (NULL == pathinfo || !S_ISREG(sce->st.st_mode)) {
-			/* no it really doesn't exists */
+		if (!r->conf.http_pathinfo
+		    || NULL == (sce = http_response_physical_pathinfo(r)))
 			return http_response_physical_path_error(r, 404, "-- file not found");
-		}
-		/* note: historical behavior checks S_ISREG() above, permitting
-		 * path-info only on regular files, not dirs or special files */
-
-		/* we have a PATHINFO */
-		if (pathinfo) {
-			size_t len = r->physical.path.ptr+pathused-1-pathinfo, reqlen;
-			if (r->conf.force_lowercase_filenames
-			    && len <= (reqlen = buffer_clen(&r->target))
-			    && buffer_eq_icase_ssn(r->target.ptr + reqlen - len, pathinfo, len)) {
-				/* attempt to preserve case-insensitive PATH_INFO
-				 * (works in common case where mod_alias, mod_magnet, and other modules
-				 *  have not modified the PATH_INFO portion of request URI, or did so
-				 *  with exactly the PATH_INFO desired) */
-				buffer_copy_string_len(&r->pathinfo, r->target.ptr + reqlen - len, len);
-			} else {
-				buffer_copy_string_len(&r->pathinfo, pathinfo, len);
-			}
-
-			/*
-			 * shorten uri.path
-			 */
-
-			buffer_truncate(&r->uri.path, buffer_clen(&r->uri.path) - len);
-			buffer_truncate(&r->physical.path, (size_t)(pathinfo - r->physical.path.ptr));
-		}
 	}
 
 	if (!r->conf.follow_symlink
@@ -350,8 +233,11 @@ static handler_t http_response_config (request_st * const r) {
 
     /* do we have to downgrade from 1.1 to 1.0 ? (ignore for HTTP/2) */
     if (__builtin_expect( (!r->conf.allow_http11), 0)
-        && r->http_version == HTTP_VERSION_1_1)
+        && r->http_version == HTTP_VERSION_1_1) {
         r->http_version = HTTP_VERSION_1_0;
+        /*(when forcing HTTP/1.0, ignore (unlikely) Connection: keep-alive)*/
+        r->keep_alive = 0;
+    }
 
     if (__builtin_expect( (r->reqbody_length > 0), 0)
         && 0 != r->conf.max_request_size   /* r->conf.max_request_size in kB */
@@ -372,31 +258,38 @@ http_response_prepare (request_st * const r)
 {
     handler_t rc;
 
-	/* looks like someone has already made a decision */
+	/* abort processing if error status, e.g. while parsing request hdrs */
 	if (__builtin_expect( (r->http_status > 200), 0)) { /* yes, > 200 */
+		/*(since this func no longer runs subrequest_handler,
+		 * status code check could be >= 400 for idiom where
+		 * r->http_status is set and r->handler_module is set NULL
+		 * to set up for error doc handler)*/
 		if (0 == r->resp_body_finished)
 			http_response_body_clear(r, 0);
 		return HANDLER_FINISHED;
 	}
 
-	/* no decision yet, build conf->filename */
-	if (buffer_is_unset(&r->physical.path)) {
+	/* initial request processing and following HANDLER_COMEBACK generally
+	 * should reprocess the request, including resetting config, but this
+	 * might be skipped after plugins have been run and path is set,
+	 * e.g. for gw_backend authorizer mode where gw_backend docroot is set
+	 * and plugin sets up handling in subrequest_handler and then returns
+	 * HANDLER_COMEBACK after auth.
+	 * (elide reprocessing request following gw_backend authorizer when
+	 *  gw_backend sets physical.path to gw_backend docroot (if set) in
+	 *  gw_authorizer_ok() before calling http_response_reset()) */
+	if (__builtin_expect( (buffer_is_unset(&r->physical.path)), 1)) {
 
+		#if 0 /*(r->async_callback currently unused)*/
 		if (__builtin_expect( (!r->async_callback), 1)) {
+		#endif
 			rc = http_response_config(r);
 			if (HANDLER_GO_ON != rc) return rc;
+		#if 0 /*(r->async_callback currently unused)*/
 		}
 		else
 			r->async_callback = 0; /* reset */
-
-		/* we only come here when we have the parse the full request again
-		 *
-		 * a HANDLER_COMEBACK from mod_rewrite and mod_fastcgi might be a
-		 * problem here as mod_setenv might get called multiple times
-		 *
-		 * fastcgi-auth might lead to a COMEBACK too
-		 * fastcgi again dead server too
-		 */
+		#endif
 
 		if (r->conf.log_request_handling) {
 			log_error(r->conf.errh, __FILE__, __LINE__,
@@ -422,7 +315,8 @@ http_response_prepare (request_st * const r)
 		    && r->uri.path.ptr[0] == '*' && r->uri.path.ptr[1] == '\0')
 			return http_response_prepare_options_star(r);
 
-		if (__builtin_expect( (r->http_method == HTTP_METHOD_CONNECT), 0))
+		if (__builtin_expect( (r->http_method == HTTP_METHOD_CONNECT), 0)
+		    && (r->handler_module || !r->h2_connect_ext))
 			return http_response_prepare_connect(r);
 
 
@@ -440,8 +334,13 @@ http_response_prepare (request_st * const r)
 
 
 		/* transform r->uri.path to r->physical.rel_path (relative file path) */
-		buffer_copy_buffer(&r->physical.rel_path, &r->uri.path);
-#if defined(__WIN32) || defined(__CYGWIN__)
+		/* (MacOS X and Windows (typically) have case-insensitive filesystems)*/
+		__builtin_expect( (!r->conf.force_lowercase_filenames), 1)
+		  ? buffer_copy_buffer(&r->physical.rel_path, &r->uri.path)
+		  : buffer_copy_string_len_lc(&r->physical.rel_path,
+		                              BUF_PTR_LEN(&r->uri.path));
+
+#if defined(_WIN32) || defined(__CYGWIN__)
 		/* strip dots from the end and spaces
 		 *
 		 * windows/dos handle those filenames as the same file
@@ -469,11 +368,6 @@ http_response_prepare (request_st * const r)
 			buffer_truncate(b, len);
 		}
 #endif
-		/* MacOS X and Windows (typically) case-insensitive filesystems */
-		if (r->conf.force_lowercase_filenames) {
-			buffer_to_lower(&r->physical.rel_path);
-		}
-
 
 		/* compose physical filename: physical.path = doc_root + rel_path */
 		if (buffer_is_unset(&r->physical.doc_root))
@@ -502,17 +396,10 @@ http_response_prepare (request_st * const r)
 
 	if (NULL != r->handler_module) return HANDLER_GO_ON;
 
-	/*
-	 * No module grabbed the request yet (like mod_access)
-	 *
-	 * Go on and check if the file exists at all
-	 */
-
+		/* check if r->physical.path exists in the filesystem */
 		rc = http_response_physical_path_check(r);
 		if (HANDLER_GO_ON != rc) return rc;
 
-		/* r->physical.path is non-empty and exists in the filesystem */
-
 		if (r->conf.log_request_handling) {
 			log_error(r->conf.errh, __FILE__, __LINE__,
 			  "-- handling subrequest");
@@ -521,31 +408,32 @@ http_response_prepare (request_st * const r)
 			log_error(r->conf.errh, __FILE__, __LINE__,
 			  "URI          : %s", r->uri.path.ptr);
 			log_error(r->conf.errh, __FILE__, __LINE__,
-			  "Pathinfo     : %s", r->pathinfo.ptr
-			                     ? r->pathinfo.ptr
-			                     : "");
+			  "Pathinfo     : %.*s",
+			  BUFFER_INTLEN_PTR(&r->pathinfo));
 		}
 
-		/* call the handlers */
+		/* request handler selection */
 		rc = plugins_call_handle_subrequest_start(r);
 		if (HANDLER_GO_ON != rc) return rc;
 
-		if (__builtin_expect( (NULL == r->handler_module), 0)) {
+		if (NULL != r->handler_module) return HANDLER_GO_ON;
+
 			/* no handler; finish request */
 			if (__builtin_expect( (0 == r->http_status), 0)) {
 				if (r->http_method == HTTP_METHOD_OPTIONS) {
 					http_response_body_clear(r, 0);
 					http_response_prepare_options_star(r); /*(treat like "*")*/
 				}
-				else if (!http_method_get_head_post(r->http_method))
+				else if (r->http_method == HTTP_METHOD_CONNECT)
+					/* 405 Method Not Allowed */
+					return http_status_set_error_close(r, 405);
+					/*return http_response_prepare_connect(r);*/
+				else if (!http_method_get_head_query_post(r->http_method))
 					r->http_status = 501;
 				else
 					r->http_status = 403;
 			}
 			return HANDLER_FINISHED;
-		}
-
-		return HANDLER_GO_ON;
 }
 
 
@@ -719,12 +607,17 @@ http_response_write_prepare(request_st * const r)
         break;
       case 204: /* class: header only */
       case 205:
-      case 304:
-        /* disable chunked encoding again as we have no body */
-        http_response_body_clear(r, 1);
-        /* no Content-Body, no Content-Length */
+        /* RFC9110
+         * https://www.rfc-editor.org/rfc/rfc9110#name-content-length
+         *  A server MUST NOT send a Content-Length header field in any response
+         *  with a status code of 1xx (Informational) or 204 (No Content)
+         * (done for 205, too, only as sanity check; 205 has no content.
+         *  Content-Length: 0 will subsequently be set for 205 further below) */
         http_header_response_unset(r, HTTP_HEADER_CONTENT_LENGTH,
                                    CONST_STR_LEN("Content-Length"));
+        __attribute_fallthrough__
+      case 304: /* cooperate with http_response_304() */
+        http_response_body_clear(r, 1);
         r->resp_body_finished = 1;
         break;
       default: /* class: header + body */
@@ -743,14 +636,13 @@ http_response_write_prepare(request_st * const r)
       case HANDLER_FINISHED:
         break;
       default:
-        log_error(r->conf.errh, __FILE__, __LINE__,
-          "response_start plugin failed");
-        return HANDLER_ERROR;
+        return HANDLER_ERROR; /*(unexpected; plugin mis-coded)*/
     }
 
     if (r->resp_body_finished) {
         /* check for Range request (current impl requires resp_body_finished) */
-        if (r->conf.range_requests && http_range_rfc7233(r) >= 400)
+        if (r->conf.range_requests && r->http_status == 200
+            && http_range_rfc7233(r) >= 400)
             http_response_static_errdoc(r); /* 416 Range Not Satisfiable */
 
         /* set content-length if length is known and not already set */
@@ -785,15 +677,15 @@ http_response_write_prepare(request_st * const r)
             }
         }
     }
-    else if (r->http_version == HTTP_VERSION_2) {
-        /* handled by HTTP/2 framing */
+    else if (r->http_version >= HTTP_VERSION_2) {
+        /* handled by HTTP/2 or HTTP/3 framing */
     }
     else {
         /**
          * response is not yet finished, but we have all headers
          *
          * keep-alive requires one of:
-         * - Content-Length: ... (HTTP/1.0 and HTTP/1.0)
+         * - Content-Length: ... (HTTP/1.1 and HTTP/1.0)
          * - Transfer-Encoding: chunked (HTTP/1.1)
          * - Upgrade: ... (lighttpd then acts as transparent proxy)
          */
@@ -881,13 +773,8 @@ http_response_call_error_handler (request_st * const r, const buffer * const err
             chunkqueue_reset(&r->reqbody_queue);
         }
 
-        r->con->is_writable = 1;
-        r->resp_body_finished = 0;
-        r->resp_body_started = 0;
-
         r->error_handler_saved_status = r->http_status;
         r->error_handler_saved_method = r->http_method;
-
         r->http_method = HTTP_METHOD_GET;
     }
     else { /*(preserve behavior for server.error-handler-404)*/
@@ -901,6 +788,11 @@ http_response_call_error_handler (request_st * const r, const buffer * const err
     buffer_copy_buffer(&r->target, error_handler);
     http_response_errdoc_init(r);
     r->http_status = 0; /*(after http_response_errdoc_init())*/
+    /* paranoia; mistake if error handler does not ignore "upgrade" */
+    if (light_btst(r->rqst_htags, HTTP_HEADER_UPGRADE))
+        http_header_request_unset(r, HTTP_HEADER_UPGRADE,
+                                  CONST_STR_LEN("upgrade"));
+    r->h2_connect_ext = 0;
     return 1;
 }
 
